//#################################################################################################//
//#                                                                                               #// 
//#  **** Schneider Engineering, Bad Vilbel ***                                                   #//
//#  **** Inh.  Dipl.-Ing. Peter Schneider  ***                                                   #//
//#                                                                                               #//
//#  Filename : astarLibrary.h                                                                    #//
//#  Project  : Odemetrie                                                                         #//
//#                                                                                               #//
//#                                                                                               #//    
//#  Odemetrie mit A* Pathfindig Algorithmus ein moeglicher Implementierungsvorschlag             #// 
//#  Quelle des Algorithmus www.wikipedia.de                                                      #//
//#                                                                                               #//
//#  (c) Peter Schneider (Schneider Engineering software / hardware development)                  #//
//#  www.ps-robotics.com                                                                          #//
//#                                                                                               #//
//#  Version:                                                                                     #//         
//#  1.00       10 July 2011 07:55       Original release                                         #//
//#                                                                                               #//
//# This program is free software; you can redistribute it and/or modify                          #//
//# it under the terms of the GNU General Public License as published by                          #//
//# the Free Software Foundation; either version 2 of the License, or                             #//
//# (at your option) any later version.                                                           #//
//#                                                                                               #//		
//# This program is distributed in the hope that it will be useful,                               #//
//# but WITHOUT ANY WARRANTY; without even the implied warranty of                                #//	
//# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                                 #//	
//# GNU General Public License for more details.                                                  #//
//#                                                                                               #//
//# You should have received a copy of the GNU General Public License                             #//
//# along with this program; if not, write to peter.schneider@schneider-engineerings.de           #//
//#                                                                                               #//
//#################################################################################################//
//
//
//          G1         G4         G6                 G1 - G8  zeigen / pruefen die Umgebung 
//       x-1, y-1    x, y-1    x+1, y-1 
//
//          G2          X         G7                 um den derzeitigen Cursor Standpunkt
//        x-1, y                x+1, y
//
//          G3         G5         G8
//        x-1, y+1    x, y+1    x+1, y+1
//
/*
//            10   
//         14  ^  14         Gewichtung G der moeglichen Richtungen
//           \   /
//       10<   X   >10
//           /   \
//         14  v  14
//            10
*/
//
// Der A*-Algorithmus 
// 
// Der A*-Algorithmus in seiner allgemeinen Form ist ein Suchalgorithmus fr den krzesten
// Pfad in kantengewichteten Graphen (mit nicht-negativen Gewichten). Als heuristischer
// Suchalgorithmus ist die Grundlage von A* die heuristische Funktion H(n).
// Diese schtzt die minimale Entfernung vom Knoten n zum Zielknoten, liefert also
// eine untere Schranke fr den noch zurckzulegendenWeg. 
// Weiter gibt es eine Funktion G(n), welche die Kosten fr den Weg vom Startknoten zum Knoten n
// liefert. Dabei bestehen die Kosten aus der Entfernung und evtl. unterschiedlichen Kosten fr 
// unterschiedliches Gelnde. 
// Damit folgt als Gesamtkostenfunktion F(n) fr einen Knoten n: F(n) = G(n) + H(n).
// 
// Der Schlssel dazu, welche Quadrate fr den Pfad in Frage kommen, ist 
// folgende Gleichung:  F = G + H 
// wobei 
//  - G = Die Bewegungskosten, um vom Startpunkt A zu einem gegebenen Quadrat des Gitters unter 
//    Verwendung des dafr ermittelten Pfades zu gelangen. 
//    In diesem Beispiel weisen wir jedem in horizontaler
//    oder vertikaler Richtung beschrittenen Quadrat einen Kostenwert von 10 und jedem diagonal 
//    beschrittenen Quadrat einen Kostenwert von 14 zu. Wir verwenden diese Werte, da sich die 
//    (abgerundete) Lnge der Diagonale in dem Quadrat nach dem Satz des Pythagoras 
//    ergibt: 14,1421.. = Quadratwurzel aus (102 + 102)). Wir verwenden 10 und 14 der Einfachheit 
//    halber. Dies hat die Vorteile, dass wir nicht Wurzeln und Dezimalzahlen berechnen mssen
//    und dennoch das Verhltnis zwischen vertikalen bzw. horizontalen und diagonalen Bewegungskosten 
//    genau genug ist. 
//
//  - H = Die geschtzten Kosten, um von dem gegebenen Quadrat zum Zielpunkt B zu gelangen. 
//    Dies wird oft heuristisch genannt, was ein bisschen verwirrend sein kann. Der Grund, 
//    warum dies so genannt wird, ist, dass diese Kosten auf Vermutung beruhen, denn tatschlich
//    kennen wir die wirkliche Entfernung erst, wenn wir den Pfad dorthin gefunden und auf dem Weg 
//    liegende Hindernisse (Wnde, Wasser, etc.) bercksichtigt haben.  
//    Wir erzeugen unseren Pfad, indem wir wiederholt unsere offene Liste durchschreiten und das 
//    Quadrat mit den geringsten F-Kosten whlen.  
// 
// Da G entlang eines spezifischen Pfades zu einem gegebenen Quadrat berechnet wird, ermitteln 
// man G fr dieses Quadrat, indem die G-Kosten eines Vorgngerquadrates genommen werden und je nach
// Bewegungsrichtung 10 oder 14 addiert wird. Der Nutzen dieser Methode erffnet sich ein bisschen spter 
// in diesem Beispiel, wenn das Objekt sich mehr als ein Quadrat vom Startquadrat fortbewegen. 
// 
// H kann auf vielerlei Art geschtzt werden. Die Methode, die hier verwendet wird, wird als 
// "Manhattan-Methode" bezeichnet, mit der die Gesamtzahl von Quadraten, die in horizontaler 
// oder vertikaler Richtung beschritten werden mssen, berechnet werden. Dabei werden diagonale 
// Bewegungen und jedwede Hindernisse auer Acht gelassen. Man multipliziert die Summe dann 
// mit 10. Dies wird (mglicherweise) die Manhattan-Methode genannt, weil sie dem Zhlen von
// Stadtvierteln gleicht, wenn man von einem Platz zu einem anderen mchte und es keine diagonale 
// Verbindung gibt. 
// 
// Zuletzt bentigt man noch zwei Listen A und B, in denen man die noch zu untersuchenden
// Knoten (A) bzw. die bereits abgearbeiteten Knoten (B) speichert.
// 
// 
// Algorithmus
// 1) Fge das Startquadrat der offenen Liste hinzu. 
// 
// 2) Wiederhole das Folgende: 
//    a) Suche in der offenen Liste nach dem Quadrat mit dem niedrigsten F-Wert. Wir bezeichnen 
//       dieses Quadrat im Folgenden als das aktuelle Quadrat. 
// 
//    b) Verschiebe es in die geschlossene Liste. 
// 
//    c) Fr jedes der 8 an das aktuelle Quadrat angrenzenden Quadrate: 
//       - Wenn es nicht begehbar ist oder sich bereits in der geschlossenen Liste befindet,
//         ignoriere es; andernfalls mach das Folgende:
// 
//       - Wenn es nicht in der offenen Liste ist, fge es der offenen Liste hinzu. Trage das
//         aktuelle Quadrat als Vorgngerquadrat dieses Quadrats ein. Trage zustzlich die Werte 
//         fr die F-, G- und H-Kosten dieses Quadrates ein.
// 
//       - Falls es bereits in der offenen Liste ist, prfe, ob der Pfad vom aktuellen Quadrat 
//         zu ihm - gemessen am G-Wert -, besser ist, als der Pfad von seinem eingetragenen  
//         Vorgngerquadrat (ein geringerer G-Wert bedeutet einen besseren Pfad). Falls dem so ist,
//         ndere sein Vorgngerquadrat auf das aktuelle Quadrat und berechne seine Werte 
//         fr G und F neu. Sofern die offene Liste nach dem F-Wert sortiert ist, ist 
//         mglicherweise eine Neusortierung dieser Liste erforderlich, um dieser Vernderung 
//         Rechnung zu tragen.
// 
//    d) Beende den Prozess, falls: 
// 
//       - das Zielquadrat in die geschlossene Liste verschoben ist; in diesem Fall wurde der Pfad ermittelt
// 
//       - kein Zielquadrat gefunden werden konnte und die offene Liste leer ist; in diesem Fall gibt es
//         keinen Pfad. Dieser Fall muss abgefangen werden, sonst luft die geschlossen Liste (A) ber
// 
// 3) Sichere den Pfad. Der Pfad erschliet sich, indem man vom Zielquadrat aus Quadrat fr Quadrat rckwrts
//    schreitend das Startquadrat erreicht. 
// 
//    - Nehme den letzten Knoten der geschlossenen Liste (B), dieser entspricht dem Zielknoten. 
// 
//    - Der letzte Knoten der Liste (B) enthlt als information den Vorgngerknoten. Suche diesen 
//      Knoten in der Liste (B)
// 
//    - Der nun gefundenen Knoten in der Liste (B) enthlt wiederum die Info des Vorgngerknotens
// 
//    - Wiederhole diesen Vorgang bis Du zum ersten Knoten in der Liste (B) kommst, dass ist der Startknoten 
// 
//
// **** benoetigte header Dateien ****
#include <stdio.h>
#include <conio.h>
#include <stdlib.h>
#include <ctype.h>

#include "astarDefaults.h"



// ***********  Globale Funktionen ***************
// *****
void  init_hindernis (void);
void  minsort (void);
void  ZeichneWeg (void);
void  astar (int, int, int, int);





// *************  hindernis und RAHMEN zeichnen ************* 
// *****
void init_hindernis (void)
{ 

   int i;
   int j;

   for  (i = 0; i <= xh; i++)                       // Init Matrix  alle Werte auf "0"
   {
      for  (j = 0; j <= yh; j++)
         matrix [i][j] = 0;
   }   

   clrscr ();                                       // Clear Screen
   
   matrix [4][2]   = 1;                             // Werte in der Umbegungsmatrix setzen
   matrix [4][3]   = 1;                             // der Wert "1"  bedeutet Hindernis 
   matrix [4][4]   = 1;                             // durch die Matrix Init Funktion sind alle
  
   gotoxy (4+1,2+1);                                // Zeichenen des Hindernisses auf dem Bildschirm
   printf ("#");
   gotoxy (4+1,3+1);         
   printf ("#");
   gotoxy (4+1,4+1);       
   printf ("#");


   matrix [8][7]   = 1;                             // durch die Matrix Init Funktion sind alle
   matrix [8][8]   = 1;                             // anderen werte  "0" 
   matrix [8][9]   = 1;                             // durch die Matrix Init Funktion sind alle
   matrix [8][10]  = 1;           
   matrix [8][11]  = 1;
   matrix [8][12]  = 1;
   matrix [8][13]  = 1;
   gotoxy (8+1,7+1);         
   printf ("#");                  
   gotoxy (8+1,8+1);                 
   printf ("#");                 
   gotoxy (8+1,9+1);         
   printf ("#");  
   gotoxy (8+1,10+1);         
   printf ("#");
   gotoxy (8+1,11+1);         
   printf ("#");
   gotoxy (8+1,12+1);         
   printf ("#");
   gotoxy (8+1,13+1);       
   printf ("#");
   
   

   matrix [20][4]  = 1;
   matrix [20][5]  = 1;
   matrix [20][6]  = 1;
   matrix [20][7]  = 1;
   matrix [20][8]  = 1;
   matrix [20][9]  = 1;
   matrix [20][10]  = 1;
   matrix [20][11]  = 1;
   matrix [20][12]  = 1;
   matrix [20][13]  = 1;
      matrix [19][13]  = 1;
      matrix [18][13]  = 1;
      matrix [17][13]  = 1;
      matrix [16][13]  = 1;
      matrix [15][13]  = 1;
      matrix [14][13]  = 1;
      
   gotoxy (20+1,4+1);         
   printf ("#");
   gotoxy (20+1,5+1);       
   printf ("#");
   gotoxy (20+1,6+1);         
   printf ("#");
   gotoxy (20+1,7+1);       
   printf ("#");
   gotoxy (20+1,8+1);       
   printf ("#");
   gotoxy (20+1,9+1);         
   printf ("#");
   gotoxy (20+1,10+1);       
   printf ("#");   
   gotoxy (20+1,11+1);         
   printf ("#");
   gotoxy (20+1,12+1);       
   printf ("#");
   gotoxy (20+1,13+1);         
   printf ("#");
     gotoxy (19+1,13+1);         
     printf ("#");
     gotoxy (18+1,13+1);         
     printf ("#");
     gotoxy (17+1,13+1);         
     printf ("#");
     gotoxy (16+1,13+1);         
     printf ("#");
     gotoxy (15+1,13+1);         
     printf ("#");
     gotoxy (14+1,13+1);         
     printf ("#");   
   
   
   
      matrix [29][8]  = 1;
      matrix [30][8]  = 1;
      matrix [31][8]  = 1;
   matrix [29][9]   = 1;
   matrix [29][10]  = 1;
   matrix [29][11]  = 1;  
      matrix [29][12]   = 1;
      matrix [30][12]  = 1;
      matrix [31][12]  = 1;
   matrix [31][9]   = 1;
   matrix [31][10]  = 1;

   
      gotoxy (29+1,8+1);         
      printf ("#");
      gotoxy (30+1,8+1);         
      printf ("#");
      gotoxy (31+1,8+1);         
      printf ("#");      
   gotoxy (29+1,9+1);         
   printf ("#");
   gotoxy (29+1,10+1);         
   printf ("#");
   gotoxy (29+1,11+1);         
   printf ("#");
      gotoxy (29+1,12+1);         
      printf ("#");
      gotoxy (30+1,12+1);         
      printf ("#");
      gotoxy (31+1,12+1);         
      printf ("#");
   gotoxy (31+1,9+1);         
   printf ("#");
   gotoxy (31+1,10+1);         
   printf ("#");

   
         
      
   // Begrenzung setzen
   for (i=0; i <xh+1 ;i++)
   {
      matrix [i][0]     = 1; 
      matrix [i][yh+1]  = 1;
   }
      
   for (i=0; i <yh+1; i++)
   {
      matrix [0][i]     = 1; 
      matrix [xh+1][i]  = 1;
   } 


   // ************ Rahmen zeichnen ************* 
   for  (i=xl;i<=(xh+1);i++)
   {
           gotoxy (i,(yl-1));
           printf ("-");
           gotoxy (i,(yh+2));
           printf ("-");
   }
   for  (i=yl;i<=(yh+1);i++)
   {
           gotoxy ((xl-1),i);
           printf ("|");
           gotoxy ((xh+2),i);
           printf ("|");
   }

   gotoxy (Start_X+1, Start_Y+1);                      // Start und Ziel zeichnen
   printf ("S");
        
   gotoxy (Ziel_X+1, Ziel_Y+1);
   printf ("Z");  

}  // Ende Funktion init_hindernis
// ------------------------------------------------------------------------------------------------------











// ------------------------------------------------------------------------------------------------------------------
// ******************* Funktion ZeichneWeg ***********************
// Zeichne Weg in die Matrix
void ZeichneWeg (void)
{
   int   path;
   int   i;

   path = listeB[index_B-1].Vorg;
            
   gotoxy (6+xh, 3);
   printf ("Knoten= %d", listeB[index_B-1].KNr);   
             
   for ( i = index_B-1;  i >= 0;  i--)
   {  
      if (path == listeB[i].KNr)
      {
           path = listeB[i].Vorg;
                                   
           // gotoxy (65,3 + m);
           // printf ("Knoten= %d", listeB[i].KNr);
           // m++;
                  
           gotoxy (listeB[i].X+1, listeB[i].Y+1);
           printf ("o");
       }   // Ende IF 
   }  // Ende For-Schleife
}  // Ende Funktion  ZeichneWeg
// ------------------------------------------------------------------------------------------------------------------











    


// ------------------------------------------------------------------------------------------------------
// *************************  Funktion MINSORT ***********************
// Sortiere eine Liste, so das der kleinster Wert der Liste, vorne in der Liste steht
// alle folgenden Elemente sind ebenfalls Aufsteigend nach ihrer Groesse sortiert
//
// Algorithmus
// Das kleinste Element gesucht und an die erste Stelle gesetzt. Danach wird aus den verbleibenden wieder
// das kleinste Element gesucht und an die zweite Stelle gesetzt. Dies geschieht bis alle Stellen gesetzt 
// sind und somit das Array sortiert ist.
//
void minsort (void)
{
   int      n1    = yh*xh+2;                         // hier die Groesse der Liste angeben
   int      links = 1;                               // erstes Element (links) in der Liste
   int      i;                                       // Zaehlvariable
   int      min;                                     // Minimum 
   int      temp;                                    // temp Variable

   while (links < n1)
   { 
      min = links;
      for (i = links; i < n1; i++)
      {
         if (( listeA[i].F < listeA[min].F) && ( listeA[i].G != 0 )) 
            min = i;                                 // Sonderfall das Element mit der Wertigkeit "0"
      }                                              // nicht beachten

      temp               = listeA[min].H;            // Vertausche links mit min aber nur wenn min 
      listeA[min].H      = listeA[links].H;          // nicht  "0"  ist
      listeA[links].H    = temp;   
            
      temp               = listeA[min].X;               
      listeA[min].X      = listeA[links].X;             
      listeA[links].X    = temp;  
            
      temp               = listeA[min].Y;            
      listeA[min].Y      = listeA[links].Y; 
      listeA[links].Y    = temp;  
            
      temp               = listeA[min].KNr;            
      listeA[min].KNr    = listeA[links].KNr; 
      listeA[links].KNr  = temp;  
            
      temp               = listeA[min].Vorg;            
      listeA[min].Vorg   = listeA[links].Vorg; 
      listeA[links].Vorg = temp;  
       
      temp               = listeA[min].G;            
      listeA[min].G      = listeA[links].G; 
      listeA[links].G    = temp;  
            
      temp               = listeA[min].F;            
      listeA[min].F      = listeA[links].F; 
      listeA[links].F    = temp;       
               
      links++;
   }  // Ende While-Schleife
}   // Ende Funktion  
// ------------------------------------------------------------------------------------------------------      


  
      
    






// -----------------------------------------------------------------------------------------------------------
// G = Bewegungskosten  diagolnal = 14    horizontal/vertical = 10
// F = Anzahl der Knoten bis Ziel ohne auf Hindernisse zu achten
// F = (Xziel - Xist) + (Yziel - Yist) entspricht der Manhattan Methode (es gibt keine Diagoinalen Verbindungen)
//
void  astar(int startx, int starty, int zielx, int ziely)
{
   
   int i;
   int j;
   int k;
   int l;
   int m;              // ++++++ Es fehlt noch die Abfrage " KEINE Ergebnis gefunden --> Liste (A) leer  +++++++
   int u;          
   int H;
   int G;
   int F;
   int x;
   int y;
   int merker_liste_A;                        // merker ob element bereits in liste A   open List
   int G_Temp;
   int merker_startelement;                   // Merker das Knoten der Startknoten ist

   int X;
   int Y; 


   // *** Startwerte in den ersten Knoten eintragen ***
   listeA[index_A].X       = startx;                   
   listeA[index_A].Y       = starty;              
   listeA[index_A].H       = 0;
   listeA[index_A].F       = 5;
   listeA[index_A].G       = 0;
   listeA[index_A].KNr     = 0;
   listeA[index_A].Vorg    = 0;


   
   merker_startelement     = 0;
   merker_liste_A          = 0;
   u                       = 0;

   while (u==0)
   {
        
      // Liste A nach niedrigsten Kosten F sortieren mit der Funktion "minsort"
      minsort ();   
      
      // ---------------------------------------------------------------------------------------------------                              
      // Wert mit geringstem Kosten F aus der openListe (Liste A)
      // holen, nach dem die Funktion Minsort() gelaufen ist
      // Kopiere erstes Element aus A nach B wenn F != 0      
      if (listeA[0].F > 0)               
      {         
         listeB[index_B] = listeA[0];                   // Kopie A nach B
         index_B++;                                     // Erhhe index der Liste B
         if (index_B > (yh*xh+2)                        // Index_B pruefen, sonst laeuft das ARRAY ueber
            index_B = yh*xh+2;

         VorKnoten = listeA[0].KNr;                     // Sichern des Vorgaengerknotens
      }
      // -----------------------------------------------------------------------------------------------------

 
      
      
          
      // -----------------------------------------------------------------------------------------------------------
      // Check ob Ziel erreicht
      if ( (listeA[0].H == 0) && (listeA[0].G != 0) ) 
      {           
         listeB[index_B] = listeA[0];                    // Kopie A nach B
         index_B++;                                      // Erhhe index der Liste B
     	 
         if (index_B > (yh*xh+2)                         // Index_B pruefen, sonst laeuft das ARRAY ueber
            index_B = yh*xh+2;
         
         gotoxy (6+xh, 6);
         printf ("fertig");                              // Ziel erreicht berechne Weg und break while Schleife 
        
         u = 1;                                          // Break while schleife 
         
         
      }
      // -----------------------------------------------------------------------------------------------------------
   

      

      
      // -----------------------------------------------------------------------------------------------------------
      // Das nach Liste B kopierte Element aus der Liste A entfernen. Somit kommt an die erste Stelle der Liste A
      // ein neuer Knoten, der untersucht werden muss
      for (i=0; i<=index_A; i++)   
      { 
         listeA[i] = listeA[i+1];                        // Wenn F-Wert > 0    F = Kosten
      }
      
      
      for (i=index_A; i<=(yh+yh+1); i++)                 // Delete alle Elemnete in Liste A die > index_A
      {
         listeA[i].X    = 0;
         listeA[i].Y    = 0;
         listeA[i].F    = 0;   
         listeA[i].G    = 0;   
         listeA[i].H    = 0;       
         listeA[i].KNr  = 0;
         listeA[i].Vorg = 0;
      } 
      index_A--;                                         // Korrektur des Index wegen lschung des ersten Elementes
      if (index_A < 0)
         index_A = 0; 
      // -----------------------------------------------------------------------------------------------------------

      
          

      // -----------------------------------------------------------------------------------------------------------   
      // Entscheidung ber erstes Element, wenn Knoten gleich Startknoten ist       
      if (merker_startelement != 0)
      {
         X      = listeB[index_B-1].X;                   // X , Y der Umgebung des lestzten Knoten aus Liste B berprfen
         Y      = listeB[index_B-1].Y;           
         G_Temp = listeB[index_B-1].G;
      }
      else
      {
         X = Start_X;                                    // Da erstes Element werden hier der Startpunkt als Quelle genutzt
         Y = Start_Y;
         merker_startelement = 1;
      }
      // -----------------------------------------------------------------------------------------------------------        
   
      
      
      
      
      // -----------------------------------------------------------------------------------------------------------   
      // Umgebung des ersten Elementes der Liste A analysieren 
      // Doppelte For-Schleife um den Punkt X und Y     
      // x-1 bis x+1  und  y-1 bis y+1  
      merker_liste_A = 0;                                // merker ob element bereits in liste A   close list
      m = 0;   
                                                         // merker fr analyse der G-Werte 
      for (i = X-1; i <= X+1; i++)                       // Prfen der Umgebung der aktuellen Position
      {        										            
         for (j = Y-1; j <= Y+1; j++)
         {     
            
            // --------------------------------------------------------------------------------------------       
            // Wenn nicht Quelle gleich ueberpruefter Knoten i=X  j=Y und nicht Begrenzung / Hindernis
            // Berechnung der Werte G, H und F                
            m++;     
 
            if (!((i == X) && (j == Y) || (matrix [i] [j] == 1)) )  
            {
               
               // ------------------------------------------------------------------------------------             
               if ((m==1) || (m==3) || (m==7) || (m==9))              
                  G = 14;                                 // Diagonale ist Gewicht 14
               else 
                  G = 10;                                 // horizontale - vertikale ist Gewicht 10  
               
               // Berechnung der Kosten pro ueberprueften Knoten   
               G = G + G_Temp;                                    // G des Vorgngers hinzu addieren
               H = ( abs(zielx - i) + abs(ziely - j) ) * 10;  	  // Ausrechnen nach Manhatten Prinzip 
               F = G + H;                                         // wieviele Blcke das Ziel entfernt ist
               x = i;	
               y = j;		
               // ----------------------------------------------------------------------------------------           
         
              
               
               
               // -----------------------------------------------------------------------------------------                     
               // Pruefen ob Knoten bereits in Liste B enthalten
               for (l=0; l<=index_B; l++)              
               {                                        
                  if ((x == listeB[l].X) && (y == listeB[l].Y))   // Wenn ja setze MERKER  "merker_liste_A = 1"
                     merker_liste_A = 1;                    
               }
               // ------------------------------------------------------------------------------------------       
            
            
                      
               
               // --------------------------------------------------------------------------------------------           
               // Pruefen ob neu berechnte Knoten bereits in Liste A enthalten ist
               // Wenn der Knoten enthalten ist und die  G,F Werte schlechter sind dann mach einen Update
               // und ndere G, F  und Vorgaengerknoten auf aktuellen Vorgaengerknoten aendern
               for (l=0; l<=index_A; l++)              
               {   
                  // Check ist der Wert fuer G besser als im Knoten der Liste A                                            
                  if (((x == listeA[l].X) && (y == listeA[l].Y)) && (G < listeA[l].G))  
                  {
                     listeA[l].H    = H;                         // Update Knoten in Liste A mit 
                     listeA[l].G    = G;                         // neue Werte fuer G,H und F
                     listeA[l].F    = F;    
                     listeA[l].Vorg = VorKnoten; 	  
                 
                     merker_liste_A = 1;                         // Knoten in Liste A ueberarbeitet
                  }
                  else if ((x == listeA[l].X) && (y == listeA[l].Y))  
                  { 
                     merker_liste_A = 1;                         // Knoten in bereits in Liste A enthalten
                  }
               }  
               minsort ();                                       // Liste A nun neu Sortieren, da sich 
                                                                 // die Werte  G und F geaendert haben           	

               // ------------------------------------------------------------------------------------------------------ 
            
             
            
            
               // ------------------------------------------------------------------------------------------------------  
               // Wenn der Merker nicht gesetzt wurde, dass das Element schon in Liste A ist  dann 
               // Update Liste A  durchfuehren     
               if (merker_liste_A == 1)               
               {                             
                  merker_liste_A = 0;                             // Knoten ignorieren Merker zuruecksetzen auf Null
       	      }
          	   else
               {                                                  // Update Liste A
                  KnotenNummer++;                                 // Knotennummer immer +1 addieren

                  index_A++;                                      // Index Liste A erhoehen 
	 	  if (index_A > (yh*xh+2)                         // Index_A pruefen, sonst laeuft das ARRAY ueber
                     index_A = yh*xh+2;

                  listeA[index_A].H    = H;                       // Schreibe die ermittelten Werte an das Ende
                  listeA[index_A].G    = G;                       // der offnene Liste (a)
                  listeA[index_A].F    = F;    
                  listeA[index_A].X    = x;  
                  listeA[index_A].Y    = y;  
                  listeA[index_A].KNr  = KnotenNummer;            // laufende KnotenNr  vergeben
                  listeA[index_A].Vorg = VorKnoten;               // Vorgaenger (Ausgangsknoten der Pruefung eintragen)
               }	
               // ------------------------------------------------------------------------------------------------------  	      
            } // Ende IF  Nicht Hindernis

         }  // Ende For-Schleife  Y

      }   // Ende For-Schleife  X

              
 }  // While-Schleife         
}  // ENDE Funktion A-Star












